from gwlogging import logger
from GwErrors import GwProbeTypeError, GwProbeCommsError, \
    GwNotImplementedError, GwProbeExecutionError

from gwtimeutils import totimestamp, timedelta2seconds
from variables import PROBE_RETRY_ATTEMPTS

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
from xmodemdirk import XMODEM, NAK

#Xmodem stuff based on data from
#http://stackoverflow.com/questions/358471/xmodem-for-python


import Gw12Status
import Gw12Route
import Gw12InstantSense
import Gw12FileList

from serial import Serial
from datetime import datetime
from time import sleep
from os import makedirs
from os.path import exists

PROBE_BAUD = 9600
SSN_BAUD = 115200

WIRED_PROBE = 30
REMOTE_PROBES = range(31,35)

SSN_GATEWAY = 50
SSNS = range(52, 55)
SDNS = range(62, 65)
GEOPHONES = [SSN_GATEWAY]  + SSNS + SDNS

GEOPHONE_BROADCAST = 251
SDN_BROADCAST = 252
SSN_BROADCAST = 253
PROBE_BROADCAST = 254
COMPLETE_BROADCAST = 255

PROBE = 'probe'
GEOPHONE = 'geophone'

RS485 = 1
RADIO_868 = 2
RADIO_173 = 3
USB = 4


def device_type(node_id):
    node_id = int(node_id)
    if (node_id == COMPLETE_BROADCAST):
        logger.critical("Address reserved but not implemented")
        raise GwNotImplementedError()
    node_type = None
    if (node_id == WIRED_PROBE
            or node_id in REMOTE_PROBES
            or node_id == PROBE_BROADCAST):
        node_type = PROBE
    elif (node_id in GEOPHONES
            or node_id == GEOPHONE_BROADCAST
            or node_id == SSN_BROADCAST
            or node_id == SDN_BROADCAST):
        node_type = GEOPHONE
    else:
        raise GwProbeTypeError()
    logger.info("Node ID %d is of type %s" % (node_id, node_type))
    return node_type
def _check_geophone(node_id):
    if(device_type(node_id) != GEOPHONE):
        raise GwProbeTypeError()

class Gw12:

    version = 0.1
    probe_port_name = None
    ssn_port_name = None
    xmodemport = None

    def __init__(self, probe_port_name="/dev/probe",
            ssn_port_name="/dev/ssn"):
        self.probe_port_name = probe_port_name
        self.ssn_port_name = ssn_port_name

    def __del__(self):
        pass

    def _choose_port_(self, dest_id):
        dest_port = None
        dest_type = device_type(dest_id)
        if dest_type is PROBE:
            dest_port = Serial(self.probe_port_name, PROBE_BAUD, timeout=4)
        elif (dest_type is GEOPHONE):
            dest_port = Serial(self.ssn_port_name, SSN_BAUD, timeout=4)
            logger.info("Using SSN Serial port")
        else:
            raise GwProbeTypeError()
        return dest_port


    def _send_ssncommand_(self, cmd_string):
        """ Special case commands for wired probe"""
        port = self._choose_port_(SSN_GATEWAY)
        logger.msp("Command: '%s'" % cmd_string)
        cmd_string = cmd_string + "\r\n"
        port.flush()
        port.flushInput()
        port.write(cmd_string)
        resp = port.readlines(None)
        attempts = 1
        logger.msp(resp)
        while ((resp == [] or 'FAIL\r\n' in resp)
                and (attempts <= PROBE_RETRY_ATTEMPTS)):
            logger.warn("Attempts %i failed" % attempts)
            logger.debug("Sending new lines")
            port.flush()
            port.flushInput()
            port.write("\r\n\r\n")
            port.readlines(None)
            port.flush()
            port.flushInput()
            logger.debug("Retrying Command")
            port.write(cmd_string)
            resp = port.readlines(None)
            logger.msp(str(resp))
            attempts += 1
        if attempts > PROBE_RETRY_ATTEMPTS:
            logger.critical("Sending command failed after specified attempts")
            raise GwProbeCommsError()
        logger.debug("ARM has responded")
        port.close()
        clean = [line for line in resp if not (line.lstrip("\x00").startswith('#') or
            line.startswith('\r\n'))]
        clean.pop(0)
        for l in resp:
            if l.startswith("FAIL"):
                try:
                    logger.error("Failure code:%s" % l.split(" ")[1])
                except IndexError:
                    logger.info("No failure code returned")
                raise GwProbeExecutionError()
        if "OK\r\n" in clean:
            clean.remove("OK\r\n")
            if clean == []:
                return True
            else:
                return clean
        elif "WAIT\r\n" in clean:
            return True
        else:
            logger.error("No FAIL or OK recieved")



    def _send_wpcommand_(self, cmd_string):
        """ Special case commands for wired probe"""
        port = self._choose_port_(WIRED_PROBE)
        logger.msp("Command: '%s'" % cmd_string)
        cmd_string = cmd_string + "\r\n"
        port.flush()
        port.flushInput()
        port.write(cmd_string)
        resp = port.readlines(None)
        attempts = 1
        logger.msp(resp)
        while ((resp == [] or 'FAIL\r\n' in resp)
                and (attempts <= PROBE_RETRY_ATTEMPTS)):
            logger.warn("Attempts %i failed" % attempts)
            logger.debug("Sending new lines")
            port.flush()
            port.flushInput()
            port.write("\r\n\r\n")
            port.readlines(None)
            port.flush()
            port.flushInput()
            logger.debug("Retrying Command")
            port.write(cmd_string)
            resp = port.readlines(None)
            logger.msp(str(resp))
            attempts += 1
        if attempts > PROBE_RETRY_ATTEMPTS:
            logger.critical("Sending command failed after specified attempts")
            raise GwProbeCommsError()
        logger.debug("ARM has responded")
        port.close()
        clean = [line for line in resp if not (line.lstrip("\x00").startswith('#') or
            line.startswith('\r\n'))]
        clean.pop(0)
        for l in resp:
            if l.startswith("FAIL"):
                try:
                    logger.error("Failure code:%s" % l.split(" ")[1])
                except IndexError:
                    logger.info("No failure code returned")
                raise GwProbeExecutionError()
        if "OK\r\n" in clean:
            clean.remove("OK\r\n")
            if clean == []:
                return True
            else:
                return clean
        elif "WAIT\r\n" in clean:
            return True
        else:
            logger.error("No FAIL or OK recieved")

    def _send_command_(self, cmd_string):
        dest_id = cmd_string.split(" ")[1]
        port = self._choose_port_(dest_id)
        logger.msp("Command: '%s'" % cmd_string)
        cmd_string = cmd_string + "\r"
        port.flushInput()
        port.write(cmd_string)
        resp = port.readlines(None)
        attempts = 1
        logger.msp(resp)
        while (resp == []
                and (attempts <= PROBE_RETRY_ATTEMPTS)):
            logger.warn("Attempts %i failed" % attempts)
            logger.debug("Sending new lines")
            port.flush()
            port.flushInput()
            port.write("\r\n\r\n")
            port.readlines(None)
            port.flush()
            port.flushInput()
            logger.debug("Retrying Command")
            port.write(cmd_string)
            resp = port.readlines(None)
            logger.msp(str(resp))
            attempts += 1
        if attempts > PROBE_RETRY_ATTEMPTS:
            logger.critical("Sending command failed after specified attempts")
            raise GwProbeCommsError()
        logger.debug("ARM has responded")
        port.close()
        clean = [line for line in resp if not (line.lstrip("\x00").startswith('#') or
            line.startswith('\r\n'))]
        clean.pop(0)
        for l in resp:
            if l.startswith("FAIL"):
                try:
                    logger.error("Failure code:%s" % l.split(" ")[1])
                except IndexError:
                    logger.info("No failure code returned")
                raise GwProbeExecutionError()
        if "OK\r\n" in clean:
            clean.remove("OK\r\n")
            while "OK\r\n" in clean:
                logger.warn("multiple oks")
                clean.remove("OK\r\n")
            if clean == []:
                return True
            else:
                return clean
        elif "WAIT\r\n" in clean:
            return True
        else:
            logger.error("No FAIL or OK recieved")
            raise GwProbeExecutionError()


    def geothreshset(self, node_id, thres):
        _check_geophone(node_id)
        if thres not in range(0, 8):
            logger.critical("Not a valid thresdhold")
            raise GwError()
        logger.debug("Set threshold on %d to %d" % (node_id, thres))
        self._send_command_("geothreshset %d %d" % (node_id, thres))

    def geothreshget(self, node_id):
        _check_geophone(node_id)
        logger.debug("Getting threshold from %d" % node_id)
        return  int(self._send_command_("geothreshget %d" % node_id)[0].strip())

    def geosampleset(self, node_id, sample):
        _check_geophone(node_id)
        if sample not in range(0, 10000):
            logger.critical("Not a valid sample number")
        logger.debug("Setting sample on %d to %d" % (node_id, sample))
        self._send_command_("geosampleset %d %d" % (node_id, sample))

    def geosampleget(self, node_id):
        _check_geophone(node_id)
        logger.debug("GEtting sample from %d" % node_id)
        return int(self._send_command_("geosampleget %d" % node_id)[0].rstrip())

    def geostats(self, node_id):
        _check_geophone(node_id)
        logger.debug("Getting stats from %d" % node_id)
        return self._send_command_("geostats %d" % node_id)

    def geodisable(self, node_id):
        _check_geophone(node_id)
        self._send_command_("geodisable %d" % node_id)
        logger.warn("Geophone %d disabled" % node_id)

    def geoenable(self, node_id):
        _check_geophone(node_id)
        self._send_command_("geoenable %d" % node_id)
        logger.warn("Geophone %d enabled" % node_id)

    def geoxfer(self, node_id, dest_id):
        _check_geophone(node_id)
        logger.debug("Initiating transfer from %d to %d" % (node_id, dest_id))
        self._send_command_("geoxfer %d %d" % (node_id, dest_id))

    def filedel(self, node_id, filename):
        _check_geophone(node_id)
        # This would idedally work on probes but not implemented in time
        self._send_command_("filedel %d %s" % (node_id, filename))
        logger.warn("File %s deleted from %d" % (filename, node_id))





    def versionget(self, node_id):
        logger.warn("NOT PROPERLY IMPLEMENTED YET")
        return self._send_command_("ver %d" % node_id)[0].rstrip("\r\n")

    def nodestatus(self, node_id):
        logger.debug("Getting status from node %d" % node_id)
        resp = self._send_command_("nodestatus %d" % node_id)
        return Gw12Status.parse(resp[0])

    def nodereset(self, node_id):
        logger.warn("Resetting node %d" % node_id)
        return self._send_command_("nodereset %d" % node_id)

    def nodesleep(self, node_id):
        logger.info("Sending node %d to sleep" % node_id)
        return self._send_command_("nodesleep %d" % node_id)

    def stayawake(self, node_id, time):
        secs = timedelta2seconds(time)
        logger.info("Telling node %d to stay awake for %d seconds" % (node_id,
            secs))
        return self._send_command_("stayawake %d %d" % (node_id, secs))

    def rtcset(self, node_id, time):
        logger.debug("Setting node %d time to %s" % (node_id, time))
        return self._send_command_("rtcset %d %d" %
            (node_id, totimestamp(time)))

    def rtcget(self, node_id):
        logger.debug("Getting time from node %d" % node_id)
        resp = self._send_command_("rtcget %d" % node_id)
        logger.debug(resp)
        return datetime.utcfromtimestamp(float(resp[0].rstrip("\r\n")))

    def rtccounterset(self, node_id, counter):
        logger.debug("Getting counter on  node %d to %d" % (node_id, counter))
        return long(self._send_command_(
            "rtccounterset %d %d" % (node_id, counter)))

    def rtccounterget(self, node_id):
        logger.debug("Getting counter from node %d" % node_id)
        return long(self._send_command_("rtccounterget %d" % node_id)[0])

    def rtcalarmset(self, node_id, tme, period, duration):
        logger.debug("Setting alarm on node %d to %s, period: %s duration %s" %
            (node_id, tme, period, duration))
        return self._send_command_("rtcalarmset %d %d %d %d" %
            (node_id, totimestamp(tme), period.seconds, duration.seconds))

    def rtcalarmget(self, node_id):
        logger.debug("Getting alarm from node %d" % node_id)
        epoch = self._send_command_("rtcalarmget %d" % node_id)
        return datetime.utcfromtimestamp(epoch)

    def sensealarmset(self, node_id, tme, period):
        logger.debug("Setting sense alarm on node %d, to %s interval: node %s"
            % (node_id, tme, period))
        return self._send_command_("sensealarmset %d %d %d" % (node_id,
            totimestamp(tme), period.seconds))

    def sensstrain(self, node_id):
        logger.debug("Getting strain reading from node %d" % node_id)
        return int(self._send_command_("sensstrain %d" %
            node_id)[0].rstrip("\r\n"))

    def sensbatv(self, node_id):
        logger.debug("Getting batv reading from node %d" % node_id)
        return int(self._send_command_("sensbatv %d" %
            node_id)[0].rstrip("\r\n"))

    def senspress(self, node_id):
        logger.debug("Getting Pressure reading from node %d" % node_id)
        return int(self._send_command_("senspress %d" %
            node_id)[0].rstrip("\r\n"))

    def senstemp(self, node_id):
        logger.debug("Getting Temperature reading from node %d" % node_id)
        return int(self._send_command_("senstemp %d" %
            node_id)[0].rstrip("\r\n"))

    def senscond(self, node_id):
        logger.debug("Getting Conductivity reading from node %d" % node_id)
        return int(self._send_command_("senscond %d" %
            node_id)[0].rstrip("\r\n"))

    def sensall(self, node_id):
        logger.debug("Getting all readings from from %d" % node_id)
        return Gw12InstantSense.parse(
            self._send_command_("sensall %d" % node_id)[0])

    def sensall_raw(self, node_id):
        logger.debug("Getting all readings from from %d" % node_id)
        logger.error("This uses a screwed up version of the class")
        logger.error("Curse you, untyped ness")
        return Gw12InstantSense.parse_raw(
            self._send_command_("sensall %d" % node_id)[0])

    def idget(self, node_id):
        logger.debug("Getting id for node %d" % node_id)
        resp = self._send_command_("idget %d" % node_id)
        return resp[0].rstrip("\r\n")

    def idset(self, old_id, new_id):
        logger.debug("Changing node %d to %d" % (old_id, new_id))
        return self._send_command_("idset %d %d" % (old_id, new_id))

    def ping(self, dest_id):
        logger.debug("Pinging node:%d " % dest_id)
        try:
            if self._send_command_("ping %d" % dest_id):
                logger.info("Ping sucessful")
                return True
        except GwProbeExecutionError:
            logger.warn("Ping Failed")
            return False


    def routeset(self, node_id, route):
        logger.warn("Convinience, use route addinstead")
        self.routeadd(node_id, route)

    def routeadd(self, node_id, route):
        logger.debug("Adding the following route to node %d" % node_id)
        logger.debug(str(route))
        return self._send_command_("routeadd %d %d %d %d" %
            (node_id, route.dest_id, route.source_id, route.link))

    def routeget(self, node_id):
        logger.debug("Getting route table from %d" % node_id)
        resp = self._send_command_("routeget %d" % node_id)
        if isinstance(resp, list):
            return Gw12Route.parse(resp)
        else:
            logger.warn("Route table is empty")
            return []

    def filelist(self, node_id):
        logger.debug("Getting file list from %d" % node_id)
        if (node_id == WIRED_PROBE):
            logger.warn("Wired probe file list")
            resp = self._send_wpcommand_("ls out")
        elif (node_id == SSN_GATEWAY):
            logger.warn("ssn file list")
            resp = self._send_ssncommand_("ls out")
        else:
            resp = self._send_command_("filelist %d" % node_id)

        if isinstance(resp, list):
            return Gw12FileList.parse(resp)
        else:
            logger.warn("File list empty")
            return []

    # move a file from a node  to the wired probe
    def fileget(self, dest_id, fname):
        logger.debug("getting file %s from node:%d " % (fname, dest_id))
        try:
            if self._send_command_("fileget %d %s" % (dest_id, fname)):
                logger.info("fileget sucessful")
                return True
        except GwProbeExecutionError:
            logger.warn("fileget Failed")
            return False

    def get_files(self, node_id):
        if not exists("/glacsweb/data/out"):
            logger.warn("/glacsweb/data/out doesn't exist")
            makedirs("/glacsweb/data/out")
        logger.debug("Getting all files from %d" % node_id)
        flist = self.filelist(node_id)
        logger.msp(flist)
        for f in flist:
            fname = "out/" + f.name + "   "
            if (node_id == WIRED_PROBE
                    or node_id == SSN_GATEWAY):
                self.xm(node_id, fname)
            else:
                if self.fileget(node_id, fname):
                    # unlikely to ever get here
                    logger.info("Got file %s" % fname)
                else:
                    logger.error("Failed to get file %s " % fname)
            sleep(1)
# XMODEM STUFF

    def getc(self, size, timeout=1):#
        return self.xmodemport.read(size)

    def putc(self, data, timeout=1):
        return self.xmodemport.write(data)

    def xm(self, dest_id, filename):
        logger.debug("Getting %s" % filename)
        fullpath = '/glacsweb/data/' + filename
        filep = open(fullpath, 'wb')
        # open the serial port we use
        self.xmodemport = self._choose_port_(dest_id)
        self.xmodemport.timeout = 1
        self.xmodemport.write("\r")
        self.xmodemport.readlines(None)
        self.xmodemport.write("xm %s\r" % filename)
        # Dirk's way to clear the probe's "s_xdl"
        sleep(2)
        self.xmodemport.readlines(None)
        modem = XMODEM(self.getc, self.putc)
        logger.debug(modem.recv(filep))
        logger.debug("Download Complete")
        filep.flush()
        filep.close()
        self.xmodemport.write("\r\r")
        self.xmodemport.readlines(None)
        self.xmodemport.flush()
        self.xmodemport.close()
